// *****************************
// ***     M a t h 3 D       ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Funzioni per la matematica 3D
// con vettori e matrici

#include "math3d.h"

Vector::Vector() {
	x=y=z=0;
}

Vector::Vector(float a,float b,float c) {
	x=a;y=b;z=c;
}

void Vector::set(float a,float b,float c) {
	x=a;y=b;z=c;
}

void Vector::add(Vector &v) {
	x=x+v.x;
	y=y+v.y;
	z=z+v.z;
}

void Vector::sub(Vector &v) {
	x=x-v.x;
	y=y-v.y;
	z=z-v.z;
}

void Vector::plane(Vector &b1,Vector &b2) {
	b1.x=1.0;b1.y=1.0;
	b1.z=-(x+y)/z;
	b1.normalize();
	b2=cross(b1,*this);
}

void Vector::normalize() {
	float len=sqrt(fabs(x*x+y*y+z*z));
	x/=len;y/=len;z/=len;
}

Vector mul(float f,Vector &v) {
	return Vector(f*v.x,f*v.y,f*v.z);
}

Vector add(Vector &v1,Vector &v2) {
	return Vector(v1.x+v2.x,v1.y+v2.y,v1.z+v2.z);
}

Vector sub(Vector &v1,Vector &v2) {
	return Vector(v1.x-v2.x,v1.y-v2.y,v1.z-v2.z);
}

float length(Vector &v) {
	return sqrt(fabs(v.x*v.x+v.y*v.y+v.z*v.z));
}

float dot(Vector &v1,Vector &v2) {
	return v1.x*v2.x+v1.y*v2.y+v1.z*v2.z;
}

Vector normalize(Vector &v) {
	float len=sqrt(fabs(v.x*v.x+v.y*v.y+v.z*v.z));
	return Vector(v.x/len,v.y/len,v.z/len);
}

Vector cross(Vector &v1,Vector &v2) {
	Vector cp;
	cp.x=(v1.y*v2.z)-(v1.z*v2.y);
	cp.y+=(v1.z*v2.x)-(v1.x*v2.z);
	cp.z+=(v1.x*v2.y)-(v1.y*v2.x);
	return cp;
}

Matrix::Matrix() {
	a11=a22=a33=1;
	a12=a13=a21=a23=a31=a32=0;
}

Matrix::Matrix(Vector &v1,Vector &v2,Vector &v3) {
	a11=v1.x;a12=v2.x;a13=v3.x;
	a21=v1.y;a22=v2.y;a23=v3.y;
	a31=v1.z;a32=v2.z;a33=v3.z;
}

void Matrix::setRotX(float a) {
	a=2*M_PI*a/360;
	a11=1.0;a12=0.0;a13=0.0;
	a21=0.0;a22=cos(a);a23=-sin(a);
	a31=0.0;a32=sin(a);a33=cos(a);
}

void Matrix::setRotY(float a) {
	a=2*M_PI*a/360;
	a11=cos(a);a12=0.0;a13=sin(a);
	a21=0.0;a22=1.0;a23=0.0;
	a31=-sin(a);a32=0.0;a33=cos(a);
}

void Matrix::setRotZ(float a) {
	a=2*M_PI*a/360;
	a11=cos(a);a12=-sin(a);a13=0.0;
	a21=sin(a);a22=cos(a);a23=0.0;
	a31=0.0;a32=0.0;a33=1.0;
}

Vector Matrix::xForm(Vector &v) {
	Vector res;
	res.x=(a11*v.x)+(a12*v.y)+(a13*v.z);
	res.y=(a21*v.x)+(a22*v.y)+(a23*v.z);
	res.z=(a31*v.x)+(a32*v.y)+(a33*v.z);
	return res;
}

Matrix mul(Matrix &m1,Matrix &m2) {
	Vector row1(m1.a11,m1.a12,m1.a13);
	Vector row2(m1.a21,m1.a22,m1.a23);
	Vector row3(m1.a31,m1.a32,m1.a33);
	Vector col1(m2.a11,m2.a21,m2.a31);
	Vector col2(m2.a12,m2.a22,m2.a32);
	Vector col3(m2.a13,m2.a23,m2.a33);
	Matrix res;
	res.a11=dot(row1,col1);
	res.a12=dot(row1,col2);
	res.a13=dot(row1,col3);
	res.a21=dot(row2,col1);
	res.a22=dot(row2,col2);
	res.a23=dot(row2,col3);
	res.a31=dot(row3,col1);
	res.a32=dot(row3,col2);
	res.a33=dot(row3,col3);
	return res;
}

xyzSystem::xyzSystem() {
	wPos=Vector(0.0,0.0,0.0);
	xLoc=Vector(1.0,0.0,0.0);
	yLoc=Vector(0.0,1.0,0.0);
	zLoc=Vector(0.0,0.0,1.0);
}

void xyzSystem::reset() {
	wPos=Vector(0.0,0.0,0.0);
	xLoc=Vector(1.0,0.0,0.0);
	yLoc=Vector(0.0,1.0,0.0);
	zLoc=Vector(0.0,0.0,1.0);
}

void xyzSystem::move(float x,float y,float z) {
	wPos=add(wPos,mul(x,xLoc));
	wPos=add(wPos,mul(y,yLoc));
	wPos=add(wPos,mul(z,zLoc));
}

void xyzSystem::roll(float a) {
	Matrix m(xLoc,yLoc,zLoc);
	Matrix rz;
	rz.setRotZ(a);
	m=mul(m,rz);
	xLoc=m.xForm(Vector(1.0,0.0,0.0));
	yLoc=m.xForm(Vector(0.0,1.0,0.0));
	zLoc=m.xForm(Vector(0.0,0.0,1.0));
}

void xyzSystem::heading(float a) {
	Matrix m(xLoc,yLoc,zLoc);
	Matrix ry;
	ry.setRotY(a);
	m=mul(m,ry);
	xLoc=m.xForm(Vector(1.0,0.0,0.0));
	yLoc=m.xForm(Vector(0.0,1.0,0.0));
	zLoc=m.xForm(Vector(0.0,0.0,1.0));
	//lastRot.y+=a;
}

void xyzSystem::pitch(float a) {
	Matrix m(xLoc,yLoc,zLoc);
	Matrix rx;
	rx.setRotX(a);
	m=mul(m,rx);
	xLoc=m.xForm(Vector(1.0,0.0,0.0));
	yLoc=m.xForm(Vector(0.0,1.0,0.0));
	zLoc=m.xForm(Vector(0.0,0.0,1.0));
}

Vector xyzSystem::getEyePoint() {
	return wPos;
}

Vector xyzSystem::getViewDir() {
	return zLoc;
}

Vector xyzSystem::getUpVector() {
	return yLoc;
}

Vector xyzSystem::getSideVector() {
	return xLoc;
}

void xyzSystem::setEyePoint(Vector &v) {
	wPos=v;
}

void xyzSystem::setViewDir(Vector &v) {
	zLoc=v;
}

void xyzSystem::setUpVector(Vector &v) {
	yLoc=v;
}

void xyzSystem::normalize() {
	xLoc=cross(yLoc,zLoc);
	xLoc.normalize();
}

void xyzSystem::setAxis(Vector &xl,Vector &yl,Vector &zl) {
	xLoc=xl;
	yLoc=yl;
	zLoc=zl;
}

void xyzSystem::apply() {
	/*if (coeff<=0.0) {
		move(lastMov.x,lastMov.y,lastMov.z);
		roll(lastRot.x);
		heading(lastRot.y);
		pitch(lastRot.z);
	}
	else {
		move(sMov.x,sMov.y,sMov.z);
		roll(sRot.x);
		heading(sRot.y);
		pitch(sRot.z);
	}
	sMov=add(sMov,lastMov);
	sMov=mul(coeff,sMov);
	sRot=add(sRot,lastRot);
	sRot=mul(coeff,sRot);
	lastMov=Vector(0.0,0.0,0.0);
	lastRot=Vector(0.0,0.0,0.0);*/
	mat[0]=xLoc.x;mat[1]=xLoc.y;mat[2]=xLoc.z;mat[3]=0;
	mat[4]=yLoc.x;mat[5]=yLoc.y;mat[6]=yLoc.z;mat[7]=0;
	mat[8]=zLoc.x;mat[9]=zLoc.y;mat[10]=zLoc.z;mat[11]=0;
	mat[12]=wPos.x;mat[13]=wPos.y;mat[14]=wPos.z;mat[15]=1;
}


float sinTab[360];
float cosTab[360];
bool readyTab=false;
void setupTab() {
	if (readyTab) return;
	for(int i=0;i<360;i++) {
		sinTab[i]=sin((float)i*2*M_PI/360.0);
		cosTab[i]=cos((float)i*2*M_PI/360.0);
	}
}
