// *****************************
// ***     O b j e c t s     ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Oggetto generico e conseguenti
// specializzazioni

#ifndef OBJECTSH
#define OBJECTSH

#include "common.h"
#include "background.h"
#include "materials.h"
#include "camera.h"
#include <stdio.h>
#include <io.h>
#include <math.h>

struct LocalInfo {
	Vector lightPosObj;
	Vector lightPosElem;	
	Vector objectPos;
	float objectMat[16];
};

struct GlobalInfo {
	Vector lightPos;
	Camera *cam;
	float cameraMat[16];
	float frustum[6][4];
	float nPlane;
};
extern GlobalInfo gInfo;

struct ClipInfo {
	bool clipped;
	float dist;
};

#define MESH	0
#define LGHT	1
#define FLUI	2

#define WIREFRAME		0
#define FLAT			1
#define SMOOTH			2

// *******************************************
// *******************************************
// ****    3  D    O  B  J  E  C  T  S    ****
// *******************************************
// *******************************************


// *****************************
// ***     E l e m e n t     ***
// *****************************

class Element {
protected:
	float posx,posy,posz;
	Vector center;
	float ray;
	bool visible;
	bool antialias;
	int mode;
	LocalInfo *info;
public:
	Element();
	virtual ~Element();
	virtual void kill() = 0;
	virtual void render() = 0;
	virtual void renderFlat() = 0;
	virtual void setPosition(float,float,float) = 0;
	virtual int id() = 0;
	virtual ClipInfo clipped() = 0;
	void antialiasing(bool);
	void setMode(int);
	void setVisible(bool);
	bool getVisible();
	void setInfo(LocalInfo*);
};
typedef Element* ElementPtr;


// *****************************
// ***      O b j e c t      ***
// *****************************

class Object {
	ElementPtr *elem;
	int num;
	xyzSystem ref;
	LocalInfo info;
public:
	Object();
	~Object();
	void kill();
	void addElement(Element*);
	void transform();
	void localLight();
	void render();
	void antialiasing(bool);
	void setMode(int);
	void setVisible(bool);
	xyzSystem& system();
	Element* getElement(int);
	int getNum();
};
typedef Object* ObjectPtr;

// *****************************
// ***       W o r l d       ***
// *****************************

class World {
	ObjectPtr *obj;
	ElementPtr *list;
	Background *back;
	ClipInfo *clip;
	int num;
	int numTot;
	void quickSort(int,int);
	void setupFrustum();
public:
	World();
	~World();
	void kill();
	void addObject(Object*);
	void setBackground(Background*);
	void setPerspective(float,float,float);
	void render(Camera*);
	Object* getObject(int);
	int getNum();
};

// *****************************
// *** M e s h E l e m e n t ***
// *****************************

struct MapCoord { //anche questo math3d
	float u,v;
};

struct Triangle { //tutte definizioni che starebbero bene in un file math3d
	unsigned short a,b,c;
};

class MeshElement:public Element {
	GLuint list;
	Vector *vert;
	MapCoord *mapc;
	Triangle *face;
	Vector *norm;
	int nv,nf;
	Material mat;
	bool ready;
public:
	MeshElement();
	~MeshElement();
	void kill();
	void create(int,int,Vector*,MapCoord*,Triangle*);
	void setMaterial(Material);
	void setTransparency(bool,float);
	void setDiffuseMap(char*);
	void render();
	void renderFlat();
	void setPosition(float,float,float);
	ClipInfo clipped();
	int id();
};


// *******************************
// *** L i g h t E l e m e n t ***
// *******************************

#define DIRECTIONAL	0
#define POSITIONAL	1

struct Flare {
	Texture map;
	float pos;
	float dim;
	float rf,gf,bf;
};

class LightElement:public Element {
	Vector ldir;
	float cf,lf,qf;
	Vector lcol;
	float expon;
	int cutoff;
	float w0;
	bool vis;
	bool en1,en2,en3;
	Texture map;
	float dim;
	Flare *flare;
	int nFlare;
	bool ready2,ready3;
public:
	LightElement();
	~LightElement();
	void create(char*,float);
	void kill();
	void apply(int);
	void render();
	void renderFlat();
	void setPosition(float,float,float);
	Vector getPosition();
	void setType(int,float,Vector&,int);
	void setAttenuation(float,float,float);
	void setColor(float,float,float);
	void setFlare(char*,int);
	bool enabled();
	void enable(bool,bool,bool);
	void visible(bool);
	ClipInfo clipped();
	int id();
};


// *******************************
// *** F l u i d E l e m e n t ***
// *******************************

struct GridPoint {
	Vector pos;
	Vector norm;
	double val;
};
typedef GridPoint* GridPointPtr;

struct GridCell {
   GridPointPtr gPoint[8];
};

extern const int triTable[256][16];
extern const int edgeTable[256];

class FluidElement:public Element {
	Vector *atoms;
	Vector *speed;
	Vector *field;
	GridPoint *gridp;
	GridCell *cells;
	int gSize;
	int num;
	float mass;
	float dist;
	double iso;
	float xmin,xmax;
	float ymin,ymax;
	float zmin,zmax;
	Material mat;
	Vector gravity;
	float visc;
	bool drawAtm,drawGrd;
	bool ready;
	Vector *vList;
	Vector *nList;
	int surface(GridCell&);
	void VertexInterp(GridPoint*,GridPoint*,Vector&,Vector&);
public:
	FluidElement();
	~FluidElement();
	void create(int,int,float,float,float);
	void kill();
	void setEnvironmentMap(char*);
	void render();
	void renderFlat();
	void updateAtoms(float);
	void visAtoms(bool);
	void visGrid(bool);
	void setAmbColor(float,float,float,float);
	void setDifColor(float,float,float,float);
	void setSpcColor(float,float,float,float);
	void setEmsColor(float,float,float,float);
	void setLevel(double);
	void setGravity(Vector);
	void setMass(float);
	void setDistance(float);
	void setViscosity(float);
	void setLimits(float,float,float,float,float,float);
	void setPosition(float,float,float);
	void getAtoms(Vector*&,Vector*&,int&);
	Vector getPosition();
	ClipInfo clipped();
	int id();
};


// *******************************************
// *******************************************
// ****    2  D    O  B  J  E  C  T  S    ****
// *******************************************
// *******************************************


// *******************************
// ***    T e x Q u a d 2 D    ***
// *******************************

class TexQuad2D {
	Texture tex;
	float sizeX,sizeY;
	float posX,posY;
	float bright;
	MapCoord texC[4];
	bool ready;
	bool vis;
public:
	TexQuad2D();
	~TexQuad2D();
	void create(char*);
	void kill();
	void render();
	void setSize(float,float);
	void setPosition(float,float);
	void setTexCoord(int,float,float);
	void setBrightness(float);
	void setVisible(bool);
};


#endif
